﻿using System;
using System.IO;
using System.Linq;
using HeuristicLab.Algorithms.ALPS;
using HeuristicLab.Algorithms.GeneticAlgorithm;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis.Dynamic;
using HeuristicLab.Problems.DataAnalysis.Dynamic.ProblemInstances;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.Dynamic;
using HeuristicLab.Selection;
using Xunit;

namespace DynamicRegressionProblemDataGenerator.ExperimentGenerators {
  public class ExperimentGenerator {

    static ExperimentGenerator() {
      //ContentManager.Initialize(new PersistenceContentManager());
    }
    private const string RootDirectory = @"C:\benchmark_data";


    [Theory]
    [Trait("Generate", "Algorithm")]
    //[MemberData(nameof(Generate_GA_Mutations))]
    //[MemberData(nameof(Generate_ALPS_Mutations))]
    [MemberData(nameof(Generate_GA))]
    [MemberData(nameof(Generate_ALPS))]
    // [MemberData(nameof(Generate_GA_Fast))]
    // [MemberData(nameof(Generate_ALPS_Fast))]
    // [MemberData(nameof(Generate_GA_UltraFast))]
    // [MemberData(nameof(Generate_ALPS_UltraFast))]
    void GenerateAlgorithm<TAlgorithm>(string outputFile, Action<DynamicSymbolicRegressionProblem> parameterizeProblem, Func<TAlgorithm> createAlgorithm, Action<TAlgorithm> parameterizeAlgorithm)
      where TAlgorithm : IAlgorithm, IStorableContent
    {
      var alg = createAlgorithm();
      var problem = new DynamicSymbolicRegressionProblem();
      alg.Problem = problem;

      parameterizeProblem(problem);
      parameterizeAlgorithm(alg);
      
      ContentManager.Save(alg, Path.Combine(RootDirectory, outputFile), compressed: true);
    }
    
    [Theory]
    [Trait("Generate", "Experiment")]
    [MemberData(nameof(Generate_GA))]
    [MemberData(nameof(Generate_ALPS))]
    void GenerateMutationExperiment<TAlgorithm>(string outputFile, Action<DynamicSymbolicRegressionProblem> parameterizeProblem, Func<TAlgorithm> createAlgorithm, Action<TAlgorithm> parameterizeAlgorithm)
      where TAlgorithm : IAlgorithm, IStorableContent {
      var experiment = new Experiment();

      foreach (var mutationProbability in MutationProbabilities) {

        var alg = createAlgorithm();
        var problem = new DynamicSymbolicRegressionProblem();
        alg.Problem = problem;

        parameterizeProblem(problem);
        parameterizeAlgorithm(alg);
        if (alg.Parameters.TryGetValue("MutationProbability", out var parameter) && parameter is IValueParameter<PercentValue> mutationProbabilityParameter) {
          mutationProbabilityParameter.Value.Value = mutationProbability;
        }
        
        var batchRun = new BatchRun("Batch Run") {
          Repetitions = 50,
          Optimizer = alg
        };

        experiment.Optimizers.Add(batchRun);
      }

      ContentManager.Save(experiment, Path.Combine(RootDirectory, "Mutations", outputFile), compressed: true);
    }
    
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<GeneticAlgorithm>, Action<GeneticAlgorithm>> Generate_GA => new() {
      { "F1_GA.hl", Problem_F1(), () => new GeneticAlgorithm(), GA(100, ProportionalSelection(), 0.15, 700) },
      { "F2_GA.hl", Problem_F2(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 700) },
      { "F3_GA.hl", Problem_F3(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 700)},
      { "Friedman_Test_3_0.00_GA.hl", Problem_Friedman_Test_3_0_00(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 1_500) },
      { "Friedman_Test_6_0.00_GA.hl", Problem_Friedman_Test_6_0_00(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 2_000) },
    };
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<AlpsGeneticAlgorithm>, Action<AlpsGeneticAlgorithm>> Generate_ALPS => new() {
      { "F1_ALPS.hl", Problem_F1(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
      { "F2_ALPS.hl", Problem_F2(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
      { "F3_ALPS.hl", Problem_F3(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
      { "Friedman_Test_3_0.00_ALPS.hl", Problem_Friedman_Test_3_0_00(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 1_500) },
      { "Friedman_Test_6_0.00_ALPS.hl", Problem_Friedman_Test_6_0_00(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 2_000) },
    };

    
    public static double[] MutationProbabilities = { 0.15, 0.10, 0.05, 0.01, 0.005, 0.001, 0.0005, 0.0001, 0.00005, 0.00001, 0.0 };
    // public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<GeneticAlgorithm>, Action<GeneticAlgorithm>>[] Generate_GA_Mutations => new() {
    //   { "F1_GA.hl", Problem_F1(), () => new GeneticAlgorithm(), GA(100, ProportionalSelection(), 0.15, 700) },
    //   { "F2_GA.hl", Problem_F2(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 700) },
    //   { "F3_GA.hl", Problem_F3(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 700)},
    //   { "Friedman_Test_3_0.00_GA.hl", Problem_Friedman_Test_3_0_00(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 1_500) },
    //   { "Friedman_Test_6_0.00_GA.hl", Problem_Friedman_Test_6_0_00(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 2_000) },
    // };
    // public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<AlpsGeneticAlgorithm>, Action<AlpsGeneticAlgorithm>[]> Generate_ALPS_Mutations => new() {
    //   { "F1_ALPS.hl", Problem_F1(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
    //   { "F2_ALPS.hl", Problem_F2(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
    //   { "F3_ALPS.hl", Problem_F3(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 700) },
    //   { "Friedman_Test_3_0.00_ALPS.hl", Problem_Friedman_Test_3_0_00(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 1_500) },
    //   { "Friedman_Test_6_0.00_ALPS.hl", Problem_Friedman_Test_6_0_00(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 2_000) },
    // };
    
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<GeneticAlgorithm>, Action<GeneticAlgorithm>> Generate_GA_Fast => new() {
      { "Fast2_F1_GA.hl", Problem_F1(), () => new GeneticAlgorithm(), GA(50, ProportionalSelection(), 0.15, 700) },
      { "Fast2_F2_GA.hl", Problem_F2(), () => new GeneticAlgorithm(), GA(100, TournamentSelection(3), 0.15, 700) },
      { "Fast2_F3_GA.hl", Problem_F3(), () => new GeneticAlgorithm(), GA(250, TournamentSelection(4), 0.15, 700)},
      { "Fast3_F1_GA.hl", Problem_F1(), () => new GeneticAlgorithm(), GA(33, ProportionalSelection(), 0.15, 700) },
      { "Fast3_F2_GA.hl", Problem_F2(), () => new GeneticAlgorithm(), GA(67, TournamentSelection(3), 0.15, 700) },
      { "Fast3_F3_GA.hl", Problem_F3(), () => new GeneticAlgorithm(), GA(167, TournamentSelection(4), 0.15, 700)},
      { "Fast2_Friedman_Test_3_0.00_GA.hl", Problem_Friedman_Test_3_0_00_Speed2(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 750) },
      { "Fast2_Friedman_Test_6_0.00_GA.hl", Problem_Friedman_Test_6_0_00_Speed2(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 1_000) },
      { "Fast3_Friedman_Test_3_0.00_GA.hl", Problem_Friedman_Test_3_0_00_Speed3(), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 500) },
      { "Fast3_Friedman_Test_6_0.00_GA.hl", Problem_Friedman_Test_6_0_00_Speed3(), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 670) },
    };
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<AlpsGeneticAlgorithm>, Action<AlpsGeneticAlgorithm>> Generate_ALPS_Fast => new() {
      { "Fast2_F1_ALPS.hl", Problem_F1(), () => new AlpsGeneticAlgorithm(), ALPS(50, 10, 0.15, 700) },
      { "Fast2_F2_ALPS.hl", Problem_F2(), () => new AlpsGeneticAlgorithm(), ALPS(50, 10, 0.15, 700) },
      { "Fast2_F3_ALPS.hl", Problem_F3(), () => new AlpsGeneticAlgorithm(), ALPS(50, 10, 0.15, 700) },
      { "Fast3_F1_ALPS.hl", Problem_F1(), () => new AlpsGeneticAlgorithm(), ALPS(33, 10, 0.15, 700) },
      { "Fast3_F2_ALPS.hl", Problem_F2(), () => new AlpsGeneticAlgorithm(), ALPS(33, 10, 0.15, 700) },
      { "Fast3_F3_ALPS.hl", Problem_F3(), () => new AlpsGeneticAlgorithm(), ALPS(33, 10, 0.15, 700) },
      { "Fast2_Friedman_Test_3_0.00_ALPS.hl", Problem_Friedman_Test_3_0_00_Speed2(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 750) },
      { "Fast2_Friedman_Test_6_0.00_ALPS.hl", Problem_Friedman_Test_6_0_00_Speed2(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 1_000) },
      { "Fast3_Friedman_Test_3_0.00_ALPS.hl", Problem_Friedman_Test_3_0_00_Speed3(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 500) },
      { "Fast3_Friedman_Test_6_0.00_ALPS.hl", Problem_Friedman_Test_6_0_00_Speed3(), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 670) },
    };
    
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<GeneticAlgorithm>, Action<GeneticAlgorithm>> Generate_GA_UltraFast => new() {
      { "UltraFast_Friedman_Test_3_0.00_GA.hl", Problem_Friedman_Test_3_0_00_UltraFast(15), () => new GeneticAlgorithm(), GA(200, TournamentSelection(3), 0.15, 1_500) },
      { "UltraFast_Friedman_Test_6_0.00_GA.hl", Problem_Friedman_Test_6_0_00_UltraFast(20), () => new GeneticAlgorithm(), GA(500, TournamentSelection(4), 0.15, 2_000) },
    };
    public static TheoryData<string, Action<DynamicSymbolicRegressionProblem>, Func<AlpsGeneticAlgorithm>, Action<AlpsGeneticAlgorithm>> Generate_ALPS_UltraFast => new() {
      { "UltraFast_Friedman_Test_3_0.00_ALPS.hl", Problem_Friedman_Test_3_0_00_UltraFast(15), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 1_500) },
      { "UltraFast_Friedman_Test_6_0.00_ALPS.hl", Problem_Friedman_Test_6_0_00_UltraFast(20), () => new AlpsGeneticAlgorithm(), ALPS(100, 10, 0.15, 2_000) },
    };
    
    
    
    
    
    static Action<DynamicSymbolicRegressionProblem> Problem_F1() => Problem("F1.csv", GenerationalClock(1), 10, 25);
    static Action<DynamicSymbolicRegressionProblem> Problem_F2() => Problem("F2.csv", GenerationalClock(1), 20, 40);
    static Action<DynamicSymbolicRegressionProblem> Problem_F3() => Problem("F3.csv", GenerationalClock(1), 50, 70);
    
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_3_0_00() => Problem("Friedman_Test_3_0.00.csv", GenerationalClock(150), 35, 70);
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_6_0_00() => Problem("Friedman_Test_6_0.00.csv", GenerationalClock(200), 50, 100);
    
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_3_0_00_Speed2() => Problem("Friedman_Test_3_0.00.csv", GenerationalClock(75), 35, 70);
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_6_0_00_Speed2() => Problem("Friedman_Test_6_0.00.csv", GenerationalClock(100), 50, 100);
    
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_3_0_00_Speed3() => Problem("Friedman_Test_3_0.00.csv", GenerationalClock(50), 35, 70);
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_6_0_00_Speed3() => Problem("Friedman_Test_6_0.00.csv", GenerationalClock(67), 50, 100);
    
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_3_0_00_UltraFast(int clockInterval) => Problem("Friedman_Test_3_0.00.csv", GenerationalClock(clockInterval), 35, 70, partitionsUpdate: PartitionsUpdateMode.Wrapping);
    static Action<DynamicSymbolicRegressionProblem> Problem_Friedman_Test_6_0_00_UltraFast(int clockInterval) => Problem("Friedman_Test_6_0.00.csv", GenerationalClock(clockInterval), 50, 100, partitionsUpdate: PartitionsUpdateMode.Wrapping);

    #region Configure Problems
    static Action<DynamicSymbolicRegressionProblem> Problem(string fileName, Action<IConstrainedValueParameter<IEpochClock>> configureEpochClock, int maximumTreeDepth, int maximumTreeLength, int impactFactorRepetitionsBestSolution = 5, int impactFactorRepetitionsPopulation = 1, bool optimizeParameters = true, PartitionsUpdateMode partitionsUpdate = PartitionsUpdateMode.KeepLast) => (problem) => {
      var csvProvider = new DynSymRegCsvInstanceProvider();
      
      var problemData = csvProvider.ImportData(Path.Combine(RootDirectory, fileName));
      
      problemData.PartitionsUpdate = partitionsUpdate;
      problem.Load(problemData);
      
      configureEpochClock(problem.EpochClockParameter);
      
      problem.Grammar.GetSymbol("Exponential and Logarithmic Functions").Enabled = false;
      problem.MaximumTreeDepth = maximumTreeDepth;
      problem.MaximumTreeLength = maximumTreeLength;
      
      problem.ImpactFactorRepetitionsBestSolutionParameter.Value.Value = impactFactorRepetitionsBestSolution;
      problem.ImpactFactorRepetitionsPopulationParameter.Value.Value = impactFactorRepetitionsPopulation;
      
      problem.OptimizeParameters = optimizeParameters;
    };
    
    static Action<IConstrainedValueParameter<IEpochClock>> EvaluationClock(int interval) => (clockParameter) => {
      var clock = clockParameter.ValidValues.OfType<CountingClock>().Single(c => c.Name == "Evaluation Counting Clock");
      clock.IntervalSize = interval;
      clockParameter.Value = clock;
    };
    static Action<IConstrainedValueParameter<IEpochClock>> GenerationalClock(int interval) => (clockParameter) => {
      var clock = clockParameter.ValidValues.OfType<CountingClock>().Single(c => c.Name == "Generation Counting Clock");
      clock.IntervalSize = interval;
      clockParameter.Value = clock;
    };
    #endregion
    
    
    #region Configure Algorithms
    static Action<GeneticAlgorithm> GA(int populationSize, Action<IConstrainedValueParameter<ISelector>> configureSelector, double mutationProbability, int maximumGenerations) => (alg) => {
      alg.PopulationSize.Value = populationSize;

      configureSelector(alg.SelectorParameter);

      alg.Crossover = alg.CrossoverParameter.ValidValues.Single(c => c.Name == "SubtreeSwappingCrossover");
      
      alg.MutationProbability.Value = mutationProbability;
      alg.Mutator = alg.MutatorParameter.ValidValues.Single(m => m.Name == "MultiSymbolicExpressionTreeManipulator");

      alg.Analyzer.AddOperator(new SymbolicDataAnalysisVariableFrequencyAnalyzer());
      
      alg.MaximumGenerations.Value = maximumGenerations;
    };
    
    static Action<AlpsGeneticAlgorithm> ALPS(int layerSize, int maxLayers/*, Action<IConstrainedValueParameter<ISelector>> configureSelector*/, double mutationProbability, int maximumGenerations) => (alg) => {
      alg.PopulationSize.Value = layerSize;
      alg.NumberOfLayers.Value = maxLayers;

      //configureSelector(alg.SelectorParameter);
      
      alg.MutationProbability.Value = mutationProbability;
      alg.Mutator = alg.MutatorParameter.ValidValues.Single(m => m.Name == "MultiSymbolicExpressionTreeManipulator");

      alg.Analyzer.AddOperator(new SymbolicDataAnalysisVariableFrequencyAnalyzer());

      alg.MaximumGenerations = maximumGenerations;
    };
    
    static Action<IConstrainedValueParameter<ISelector>> ProportionalSelection() => (selectorParameter) => {
      var selector = selectorParameter.ValidValues.OfType<ProportionalSelector>().Single();
      selectorParameter.Value = selector;
    };
    static Action<IConstrainedValueParameter<ISelector>> TournamentSelection(int groupSize) => (selectorParameter) => {
      var selector = selectorParameter.ValidValues.OfType<TournamentSelector>().Single();
      selector.GroupSizeParameter.Value.Value = groupSize;
      selectorParameter.Value = selector;
    };
    #endregion
  }
}
